/// gml_DynamicArray.h --
///
///  Definition of the gml_TDynamicArray class template.
///
//  Copyright (c) 1998-2004 CLIPS-IMAG
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on August 29, 1998 (FB).

#ifndef __GMLDYNAMICARRAY__
#define __GMLDYNAMICARRAY__

#include "gml_Types.h"
#include "gml_Errors.h"

static const gml_TBlockSize    gml_cDynArrayDefaultNbElem        = 20; ///< default starting number of elements in an array
static const gml_TBlockSize    gml_cDynArrayDefaultNbIncrease    = 20; ///< default additional slots allocated on array growth


/// gml_TDynamicArray --
///    An array of fixed sized elements that can grow over time.
///    Growth is by block of elements (not one element every time).
///    Finding a free slot in the array is somewhat optimized using a
///    bitmap of used slots.

template <class T>
class gml_TDynamicArray
{
  public:

    /// Init --
    ///   Initializes the array with <nbElem> elements of size <elemSize>.
    ///   When a new element is required and the array is full, the array
    ///   is re-allocated with <nbIncrease> more elements.

    gml_TError        Init (gml_TBlockSize    nbElem      = gml_cDynArrayDefaultNbElem,
                            gml_TBlockSize    nbIncrease  = gml_cDynArrayDefaultNbIncrease);

    /// Dispose --
    ///   Release the array resources.

    void              Dispose ();

    /// Base --
    ///   Returns the base address of the array.
    ///
    ///   Warning
    ///     The base address in only valid from the Base() call until
    ///     the next call to GelElem(): GetElem() may require a relocation
    ///     of the memory if the array is full.

    T*                Base ()
                        { return fMem; }

    /// GetElem --
    ///   Get a pointer to a new element from the array.
    ///   If <index> is not NULL, <*index> recieves the index of the element in the array.
    ///
    ///   Returns
    ///     A pointer to the new element everything goes ok.
    ///     NULL if there is no more memory to grow the array.

    T*                GetElem (gml_TBlockSize* index = (gml_TBlockSize*)NULL);

    /// ReleaseElem --
    ///   Flag <elem> as not being used anymore.

    void              ReleaseElem (T* elem);



  protected:

    gml_TError        Allocate (gml_TBlockSize nbElem);

    gml_TBlockSize    fNbIncrease;  ///< size of the increased allocation when a new element is
                                    ///<  needed and the array is full

    T*                fMem;         ///< memory allocated for the array
    gml_TBlockSize    fNbElem;      ///< number of elements that the current memory can store
    gml_TBlockSize    fInUse;       ///< number of current allocated elements

    UInt32*           fBitmap;      ///< bitmap of elements in use
    gml_TBlockSize    fBMSize;      ///< size (number of 32 bit words) of used bitmap
};


#endif


