// gmlTcl_Method.cpp --
//
//    Defines the gmlTcl_TMethod class. See "gmlTcl_Method.h".
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 24, 2007 (FB).

#include "tcl.h"

#include "gmlTcl_Method.h"

#include "gml_Alloc.h"

#include "gmlTcl_Object.h"
#include "gmlTcl_Class.h"




// gmlTcl_TMethod::Init --

int gmlTcl_TMethod::Init (Tcl_Interp*     interp,
                          gmlTcl_TClass*  theClass,
                          int             objc,
                          Tcl_Obj*        objv[])
{
  int               res           = TCL_OK;
  Interp*           iPtr          = (Interp*)interp;
  char*             methodStrName = Tcl_GetString (objv[2]);
  Tcl_DString       ds;
  Tcl_Obj*          bodyPtr       = (Tcl_Obj*)NULL;
  Tcl_Obj*          argsPtr       = (Tcl_Obj*)NULL;
  Tcl_Obj*          insertv[3];
  int               argLen;

  fClass            = theClass;
  fStringName       = (char*)NULL;
  fProtectedString  = (char*)NULL;
  fProcName         = (char*)NULL;
  fRegistered       = 0;
  fProc             = (Proc*)NULL;
  fFlags            = 0;

  fBody             = (Tcl_Obj*)NULL;
  fArgList          = (Tcl_Obj*)NULL;
  fArgs             = (Tcl_Obj*)NULL;

  // Check arguments */

  if (objc != 5) {
      Tcl_WrongNumArgs(interp, 1, objv, "className methodName args body");
      return TCL_ERROR;
  }

  // Create the string name and the proc name.

  res = gmlTcl_Obj_UpdateStringNames (methodStrName, &fStringName, &fProtectedString, interp);
  if (res != TCL_OK)
    goto error;

  fProcName           = (char*)malloc (strlen (fClass->StringName ()) + strlen (fStringName) + 3);
  if (fProcName == (char*)NULL) {
    Tcl_AppendResult (interp, "could not create method \"", fStringName, "::",
                              methodStrName, "\": no more memory", (char*)NULL);
    goto error;
  }
  sprintf (fProcName, "%s::%s", fClass->StringName (), fStringName);

  // Check if the method is the constructor of the destructor.
  //  If it is the destructor, check that it doesn't take arguments.

  if (strcmp (fStringName, "constructor") == 0)
    fFlags = 1;
  else if (strcmp (fStringName, "destructor") == 0) {
    fFlags = 2;
    res   = Tcl_ListObjLength (interp, objv[3], &argLen);
    if (res != TCL_OK)
      goto error;
    if (argLen != 0) {
      Tcl_AppendResult (interp, "destructor must not accept any argument", (char*)NULL);
      res = TCL_ERROR;
      goto error;
    }
  }

  // Patch the body of the method in order to insert upvars to the class and object variables
  
  Tcl_DStringInit (&ds);
	Tcl_DStringAppend (&ds, "upvar #0 $className class $objName this\n", -1);
	Tcl_DStringAppend (&ds, Tcl_GetString (objv[4]), -1);

  bodyPtr = Tcl_NewStringObj (Tcl_DStringValue (&ds), Tcl_DStringLength (&ds));
  Tcl_IncrRefCount (bodyPtr);

  Tcl_DStringFree (&ds);

  // Patch the argument list in order to insert the objectName, className and methodName

  argsPtr = Tcl_DuplicateObj (objv[3]);
  Tcl_IncrRefCount (argsPtr);

  insertv[0] = Tcl_NewStringObj ("objName", -1);
  insertv[1] = Tcl_NewStringObj ("className", -1);
  insertv[2] = Tcl_NewStringObj ("methodName", -1);

  res = Tcl_ListObjReplace (interp, argsPtr, 0, 0, 3, insertv);
  if (res != TCL_OK)
    goto error;

  // Create the tcl "Proc" structure associated with the method.

  res = TclCreateProc (interp, iPtr->globalNsPtr, fProcName, argsPtr, bodyPtr, &fProc);
  if (res != TCL_OK)
    goto error;

  Tcl_DecrRefCount (bodyPtr);
  bodyPtr = (Tcl_Obj*)NULL;

  Tcl_DecrRefCount (argsPtr);
  argsPtr = (Tcl_Obj*)NULL;

  // Register the method to its class.

  fClass->RegisterMethod (this, fFlags);
  fRegistered = 1;

  fBody = objv[4];
  Tcl_IncrRefCount (fBody);

  fArgList = objv[3];
  Tcl_IncrRefCount (fArgList);

  return res;



error:

  if (bodyPtr != (Tcl_Obj*)NULL) {
    Tcl_DecrRefCount (bodyPtr);
    bodyPtr = (Tcl_Obj*)NULL;
  }
  if (argsPtr != (Tcl_Obj*)NULL) {
    Tcl_DecrRefCount (argsPtr);
    argsPtr = (Tcl_Obj*)NULL;
  }

  this->Dispose ();

  return TCL_ERROR;
}



// gmlTcl_TMethod::Dispose --

void gmlTcl_TMethod::Dispose ()
{
  // Free memorized body, arglist and args
  if (fBody != (Tcl_Obj*)NULL) {
    Tcl_DecrRefCount (fBody);
    fBody = (Tcl_Obj*)NULL;
  }
  if (fArgList != (Tcl_Obj*)NULL) {
    Tcl_DecrRefCount (fArgList);
    fArgList = (Tcl_Obj*)NULL;
  }
  if (fArgs != (Tcl_Obj*)NULL) {
    Tcl_DecrRefCount (fArgs);
    fArgs = (Tcl_Obj*)NULL;
  }

  // Deregister from its class.

  if (fRegistered) {
    fClass->DeregisterMethod (this, fFlags);
    fRegistered = 0;
  }

  // Free the tcl "Proc" structure associated with the method.

  if (fProc != (Proc*)NULL) {
    TclProcDeleteProc ((ClientData)fProc);
    fProc = (Proc*)NULL;
  }
  
  // Free the string names and the procedure name
  
  gml_TestAndFree (fProtectedString);
  gml_TestAndFree (fStringName);
  gml_TestAndFree (fProcName);
}






// Execute --

int gmlTcl_TMethod::Execute  (Tcl_Interp*       interp,
                              gmlTcl_TObject*   theObject,
                              Tcl_Obj*          objName,
                              Tcl_Obj*          className,
                              Tcl_Obj*          methodName,
                              int               objc,
                              Tcl_Obj*          objv[],
                              int               constructorCall)
{
  int         res             = TCL_OK;
  int         mObjc           = objc + 4;
  Tcl_Obj**   mObjv           = (Tcl_Obj**)NULL;
  int         i;
  char*       strResult;
  int         strResultLen;
  Tcl_Obj*    newResult;
  int         prefixLen;

  // Add the proc name, object name, class name and method name at the head of the
  //  argument array.

  mObjv = (Tcl_Obj**) malloc (sizeof (Tcl_Obj*) * mObjc);
  if (mObjv == (Tcl_Obj**)NULL) {
    Tcl_AppendResult (interp, "could not call method \"", fClass->StringName (),
                              "::", fStringName, "\": no more memory", (char*)NULL);
    return TCL_ERROR;
  }

  mObjv[0] = Tcl_NewStringObj (fProcName, -1);
  mObjv[1] = objName;
  mObjv[2] = className;
  mObjv[3] = methodName;
  Tcl_IncrRefCount (mObjv[0]);
  Tcl_IncrRefCount (mObjv[1]);
  Tcl_IncrRefCount (mObjv[2]);
  Tcl_IncrRefCount (mObjv[3]);

  for (i = 4; i < objc+4; i++)
    mObjv[i] = objv[i-4];

  // Do the call.

  fProc->cmdPtr = (Command*)(theObject->Command ());
  res = TclObjInterpProc ((ClientData)fProc, interp, mObjc, mObjv);

  Tcl_DecrRefCount (mObjv[0]);
  Tcl_DecrRefCount (mObjv[1]);
  Tcl_DecrRefCount (mObjv[2]);
  Tcl_DecrRefCount (mObjv[3]);

  free ((void*)mObjv);

  // Patch the result message if there was a "wrong # args" error.
  //  example
  //	    wrong # args: should be "TestClass::TestMethod objName className methodName a b c"
  //      wrong # args: should be "{a class::constructor} objName className methodName a b ?c?"
  //  is patched into
  //	    wrong # args: should be "testObject TestMethod a b c"
  //  or, at object creation (constructor call)
  //      wrong # args: should be "TestClass testObject a b c"
  
  if (res == TCL_ERROR) {

    strResult = Tcl_GetStringFromObj (Tcl_GetObjResult (interp), &strResultLen);
    if ((strResultLen >= 25) && (strncmp (strResult, "wrong # args: should be \"", 25) == 0)) {

      prefixLen = strlen (fClass->StringName ()) + strlen (fStringName) + 56;

      if (strResult[25] == '{')
        prefixLen += 2;

      if (strResultLen >= prefixLen) {
        newResult = Tcl_NewStringObj ("wrong # args: should be \"", -1);

        if (constructorCall) {
          Tcl_AppendToObj (newResult, fClass->ProtectedString (), -1);
          Tcl_AppendToObj (newResult, " ", -1);
          Tcl_AppendToObj (newResult, theObject->ProtectedString (), -1);
        } else {
          Tcl_AppendToObj (newResult, theObject->ProtectedString (), -1);
          Tcl_AppendToObj (newResult, " ", -1);
          Tcl_AppendToObj (newResult, fProtectedString, -1);
        }

        Tcl_AppendToObj (newResult, strResult + prefixLen, -1);
        Tcl_SetObjResult(interp, newResult);
      }
    }
  }

  return res;
}



// Args --

Tcl_Obj* gmlTcl_TMethod::Args ()
{
  int         res;
  int         objc;
  Tcl_Obj**   objv;
  int         i;
  int         argc;
  Tcl_Obj**   argv;

  if (fArgs != (Tcl_Obj*)NULL)
    return fArgs;

  // Build the list of arguments from the argument description list (where each
  //  element can be a list of two elements: the arg and the default value).

  res = Tcl_ListObjGetElements ((Tcl_Interp*)NULL, fArgList, &objc, &objv);
  if (res != TCL_OK) {
    fprintf (stderr, "PANIC: could not build list of argument for method %s::%s\n",
                      fClass->StringName (), fStringName);
    exit (1);
  }

  fArgs = Tcl_NewListObj (0, (Tcl_Obj**)NULL);
  for (i = 0; i < objc; i++) {
    res = Tcl_ListObjGetElements ((Tcl_Interp*)NULL, objv[i], &argc, &argv);
    if ((res != TCL_OK) || ((argc != 1) && (argc != 2))) {
      fprintf (stderr, "PANIC: could not build list of argument for method %s::%s\n",
                        fClass->StringName (), fStringName);
      exit (1);
    }
    if (argc == 1)
      Tcl_ListObjAppendElement ((Tcl_Interp*)NULL, fArgs, objv[i]);
    else
      Tcl_ListObjAppendElement ((Tcl_Interp*)NULL, fArgs, argv[0]);
  }
  Tcl_IncrRefCount (fArgs);

  return fArgs;
}
