// gmlTcl_Object.cpp --
//
//    Defines the gmlTcl_TObject class. See "gmlTcl_Object.h".
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 22, 2007 (FB).


extern "C" {
#include "tcl.h"
#include "tclInt.h"
#include "tclCompile.h"
}

#include "gmlTcl_Object.h"

#include "gmlTcl_Obj_Internal.h"


#include "gmlTcl_Entity.h"
#include "gmlTcl_Method.h"
#include "gmlTcl_Class.h"



// gmlTcl_TObject::Init --

int gmlTcl_TObject::Init (Tcl_Interp*       interp,
                          gmlTcl_TClass*    theClass,
                          int               objc,
                          Tcl_Obj*          objv[])
{
  int                 res           = TCL_OK;
  Tcl_Obj*            name          = objv[1];
  gmlTcl_TMethod*     constructor;
  Tcl_Obj*            saveInterpRes = (Tcl_Obj*)NULL;
  Tcl_Obj*            saveErrorInfo = (Tcl_Obj*)NULL;
  Tcl_Obj*            newErrorInfo  = (Tcl_Obj*)NULL;

  fClass              = theClass;
  fRegistered         = 0;
  fConstructed        = 0;


  // Call superclass initialization
  
  res = this->gmlTcl_TEntity::Init (gmlTcl_kObjectEntityType, interp, name);
  if (res != TCL_OK)
    return res;

  // Check arg count in case there is no constructor

  constructor         = fClass->Constructor ();

  if ((constructor == (gmlTcl_TMethod*)NULL) && (objc > 2)) {
    Tcl_AppendResult (interp, "too many args: there is no constructor", (char*)NULL);
    return TCL_ERROR;
  }

  // Register this object to its class

  theClass->RegisterObject (this);
  fRegistered = 1;

  // Call constructor if there is one.

  fConstructed = 1;

  if (constructor != (gmlTcl_TMethod*)NULL) {
    res = constructor->Execute (interp, this,
                                objv[1], objv[0],
                                Tcl_NewStringObj (constructor->StringName (), -1),
                                objc-2, objv+2, 1);

    if (fState == gmlTcl_kEntityState_cmdDeleted) {

      // This object command was destroyed during constructor evaluation.

      Tcl_AppendResult (interp, "object \"", fStringName,
                                "\" construction failed: constructor created an object, class or procedure with the same name",
                                (char*)NULL);
      res = TCL_ERROR;
    }

    if (res != TCL_OK) {
    
      // Save interpretor state in order to restore it after the destructor call.

      saveInterpRes = Tcl_GetObjResult (interp);
      Tcl_IncrRefCount (saveInterpRes);

      saveErrorInfo = Tcl_GetVar2Ex(interp, "errorInfo", (char*)NULL, TCL_GLOBAL_ONLY);
      Tcl_IncrRefCount (saveErrorInfo);

      goto error;
    }
  }

  fState = gmlTcl_kEntityState_normal;
  return res;

error:

  this->Dispose ();

  if (saveInterpRes != (Tcl_Obj*)NULL) {
    // There was an error on constructor call, check if we need to restore the error
    //  message in the interpreter.

    if (saveInterpRes != Tcl_GetObjResult (interp)) {
      // The destructor was called and changed the interp result.

      Tcl_SetObjResult (interp, saveInterpRes);

      newErrorInfo = Tcl_GetVar2Ex (interp, "errorInfo", (char*)NULL, TCL_GLOBAL_ONLY);

      if (saveErrorInfo != newErrorInfo) {
        // There was an error on the destructor call that changed the error info variable.

        Tcl_AppendStringsToObj (newErrorInfo, "\n\n", (char*)NULL);
        Tcl_AppendObjToObj (newErrorInfo, saveErrorInfo);
      }
    }
    Tcl_DecrRefCount (saveInterpRes);
    Tcl_DecrRefCount (saveErrorInfo);
  }

  return TCL_ERROR;
}



// gmlTcl_TObject::Dispose --

void gmlTcl_TObject::Dispose ()
{
  gmlTcl_TMethod*       destructor;

  if (fConstructed) {

    // Call destructor if there is one.

    destructor = fClass->Destructor ();
    if (destructor != (gmlTcl_TMethod*)NULL) {
      destructor->Execute  (gmlTcl_Obj_Interp (), this,
                            Tcl_NewStringObj (fStringName, -1),
                            Tcl_NewStringObj (fClass->StringName (), -1),
                            Tcl_NewStringObj (destructor->StringName (), -1),
                            0, (Tcl_Obj**)NULL);
    }

    // Remove the global variable of this object. Only do it if it was
    //  constructed, otherwise we may be deleting a global variable that existed
    //  before this object creation attempt and that caused the dispose.
  
    Tcl_UnsetVar (gmlTcl_Obj_Interp (), fStringName, TCL_GLOBAL_ONLY);
  }

  // De-register this object from its class

  if (fRegistered) {
    fClass->DeregisterObject (this);
    fRegistered = 0;
  }

  fConstructed = 0;

  // Call the superclass disposal.

  this->gmlTcl_TEntity::Dispose ();
}




// CommandHandler --

int gmlTcl_TObject::CommandHandler (Tcl_Interp*         interp,
                                    int                 objc,
                                    Tcl_Obj*            objv[])
{
  return this->MethodDispatchCmd (interp, objv[0], objc, objv);
}



// MethodDispatchCmd --

int gmlTcl_TObject::MethodDispatchCmd  (Tcl_Interp*   interp,
                                        Tcl_Obj*      objName,
                                        int           objc,
                                        Tcl_Obj*      objv[])
{
  int                   res               = TCL_OK;
  gmlTcl_TMethod*       theMethod         = (gmlTcl_TMethod*)NULL;
  char*                 methodNameStr;
  char*                 curs;
  int                   cont              = 1;
  Tcl_Obj*              classNameObj;
  Tcl_Obj*              methodNameObj;

  gmlTcl_TClass*        qualifClass;
  Tcl_Obj*              callObjName;
  gmlTcl_TObject*       callObj           = (gmlTcl_TObject*)NULL;
  Tcl_Obj*              callClassName;
  gmlTcl_TClass*        callClass         = (gmlTcl_TClass*)NULL;

  // Check arguments
  
  if (objc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", Tcl_GetString (objv[0]),
                              " <methodName> ...\"", (char*)NULL);
    return TCL_ERROR;
  }

  // Parse method name for the <className>::<methodName> form.
  
  methodNameStr = Tcl_GetString (objv[1]);
  curs = methodNameStr;
  while (cont) {
    if (*curs == 0)
      cont = 0;
    else
      if (*curs == ':')
        if (*(curs+1) == ':')
          cont = 0;
        else
          curs++;
      else
        curs++;
  }

  if (*curs != 0) {

    // The :: pattern was found in the method name, this fully qualified name.

    classNameObj = Tcl_NewStringObj (methodNameStr, curs - methodNameStr);
    Tcl_IncrRefCount (classNameObj);

    qualifClass = gmlTcl_Obj_ClassFromTclObj (classNameObj, interp);

    Tcl_DecrRefCount (classNameObj);

    if (qualifClass == (gmlTcl_TClass*)NULL)
      return TCL_ERROR;

    // Check that the specified class is an ancestor of this object's class.
    
    if (!fClass->IsAncestor (qualifClass)) {
      Tcl_AppendResult (interp, "\"", qualifClass->StringName (), "\" is not an ancestor of \"",
                                fStringName, "\"", (char*)NULL);
      return TCL_ERROR;
    }

    methodNameObj = Tcl_NewStringObj (curs+2, -1);

    theMethod = qualifClass->FindMethod (methodNameObj, interp);

    if (theMethod == (gmlTcl_TMethod*)NULL) {
      return TCL_ERROR;
    }

  } else {

    // Handle the "inherited" case.

    if (strcmp (methodNameStr, "inherited") == 0) {
    
      // Check that the object of this "inherited" call is the same as the object of the
      //  calling method.
      // Find the name of the method in the "methodName" variable in the context of the call.
    
      methodNameObj = Tcl_GetVar2Ex (interp, "methodName", (char*)NULL, 0);

      callObjName = Tcl_GetVar2Ex (interp, "objName", (char*)NULL, 0);
      if (callObjName != (Tcl_Obj*)NULL)
        callObj = gmlTcl_Obj_ObjectFromTclObj (callObjName);      

      callClassName = Tcl_GetVar2Ex (interp, "className", (char*)NULL, 0);
      if (callClassName != (Tcl_Obj*)NULL)
        callClass = gmlTcl_Obj_ClassFromTclObj (callClassName);      

      if ((callObjName == (Tcl_Obj*)NULL) ||
          (methodNameObj == (Tcl_Obj*)NULL) ||
          (callClassName == (Tcl_Obj*)NULL) ||
          (callClass == (gmlTcl_TClass*)NULL) ||
          (callObj != this)) {
        Tcl_AppendResult (interp, "\"inherited\" can only be invoked on \"this\" in the context of a method",
                                  (char*)NULL);
        return TCL_ERROR;
      }

      theMethod = callClass->FindMethod (methodNameObj, interp, 2);
      if (theMethod == (gmlTcl_TMethod*)NULL)
        return TCL_ERROR;

    } else {

      // This is a regular, unqualified, method call.      
      
      methodNameObj = objv[1];
      theMethod = fClass->FindMethod (methodNameObj, interp);
      if (theMethod == (gmlTcl_TMethod*)NULL)
        return TCL_ERROR;
    }
  }

  // Do the call

  classNameObj = Tcl_NewStringObj (theMethod->Class ()->StringName (), -1);
  Tcl_IncrRefCount (objName);
  Tcl_IncrRefCount (classNameObj);
  Tcl_IncrRefCount (methodNameObj);

  res = theMethod->Execute (interp, this,
                            objName,
                            classNameObj,
                            methodNameObj,
                            objc-2, objv+2);

  Tcl_DecrRefCount (methodNameObj);
  Tcl_DecrRefCount (classNameObj);
  Tcl_DecrRefCount (objName);

  return res;
}
